(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_disk.mli 187 2005-10-14 10:54:15Z gerd $ *)

open Umlmon_types

val disk_access : umlglobalspec -> umlspec -> jailfile -> disk_access
  (* Determines the access mode of the file for the VM. The file name
   * should have the form "/shared/<path>".
   *)

val get_disks : umlstatus -> managed_disk_file list
  (* Returns the disk status; scans for disks if necessary *)

val operation_in_progress : umlstatus -> bool
  (* Returns whether a disk operation is currently being performed *)

val kernel_arguments : umlstatus -> string list
  (* Returns the kernel arguments to pass the current disk definition. 
   *)

val lock_disks_at_startup : umlstatus -> lock_id
  (* Locks all disks used by the VM. If this is not possible, the exception
   * [Lock_failure] is raised, and no disk will get locked. [Failure] is
   * possible, too.
   *)

val proc_disk_files :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  unit ->
  (Umlmonctrl_aux.r_disks -> unit) ->
    unit
  (* Implements the [disk_files] procedure *)

val proc_disk_create :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  (string * int32 * string array) ->
  (Umlmonctrl_aux.r_simple -> unit) ->
    unit
  (* Implements the [disk_create] procedure *)

val proc_disk_create_cow :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  (string * string * string array) ->
  (Umlmonctrl_aux.r_simple -> unit) ->
    unit
  (* Implements the [disk_create_cow] procedure *)

val proc_disk_delete :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  (string * string array) ->
  (Umlmonctrl_aux.r_simple -> unit) ->
    unit
  (* Implements the [disk_delete] procedure *)

val proc_disk_copy :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  (string * string * string array) ->
  (Umlmonctrl_aux.r_simple -> unit) ->
    unit
  (* Implements the [disk_copy] procedure *)

val proc_disk_resize :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  (string * int32 * string array) ->
  (Umlmonctrl_aux.r_simple -> unit) ->
    unit
  (* Implements the [disk_resize] procedure *)

val proc_disk_options :
  Unixqueue.unix_event_system -> 
  umlstatus -> 
  Rpc_server.session ->
  unit ->
  (Umlmonctrl_aux.r_disk_options -> unit) ->
    unit
  (* Implements the [disk_options] procedure *)
