(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON-Web.
 *
 * UMLMON-Web is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON-Web is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON-Web; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: init.ml 121 2005-06-12 13:56:40Z gerd $
 * ----------------------------------------------------------------------
 *
 *)

open Wd_dialog
open Wd_types

(**********************************************************************)
(* Definition of additional output encodings                          *)
(**********************************************************************)

(* XXX *)

(**********************************************************************)
(* Database                                                           *)
(**********************************************************************)

(* XXX *)

(**********************************************************************)
(* Fatal errors                                                       *)
(**********************************************************************)



let simple_error_page (ch : Netchannels.out_obj_channel) error =
  Log.log_error (Printexc.to_string error);
  let out = ch # output_string in
  out "<html><body><h1>Internal Error</h1>\n";
  out "The application godi_admin cannot serve your request. The ";
  out "exact reason can be found in the error log. Please inform ";
  out "your administrator about this incidence.\n";
  out "</body></html>\n";
;;


(*
let extended_error_page (universe : universe_type) ch error =
  Log.log_error (Printexc.to_string error);
  (* Instantiate the template "fatal-error" of the startpage manually. *)
  (* Step 1: Create our own faked cgi_activation object *)
  let cgi_env = new Netcgi_env.custom_environment () in
  let dummy = new Netchannels.output_null () in
  cgi_env # set_output_ch dummy;
  cgi_env # set_cgi ~gateway_interface:"FAKE CGI" ~request_method:"GET" ();
  cgi_env # set_input_state `Received_header;
  cgi_env # set_output_state `Start;
  let cgi = 
    new Netcgi.std_activation ~env:(cgi_env :> Netcgi_env.cgi_environment) () in
  (* Step 2: Get the startpage object *)
  let environment = Wd_cycle.make_environment cgi in
  let startpage = universe # create environment "startpage" in
  (* Step 3: Initialize the startpage object *)
  let errid = Definitions.errid_of_error error in
  startpage # set_variable "runtime-error" (String_value errid);
(*
  let session = !Registry.new_session universe environment in
  startpage # set_variable "session" (Dialog_value (Some session));
*)
  (* Step 4: Output the template "fatal-error" *)
  let inst = Wd_template.apply_byname startpage "fatal-error" [] in
  let s = Wd_template.to_string startpage inst in
  ch # output_string s;
;;
 *)


(* [print_error] is called to output fatal errors. It is initialized with
 * a very simple definition which is replaced later by a more 
 * sophisticated one.
 *)

let print_error = ref simple_error_page
;;


(**********************************************************************)
(* Register dialogs etc.                                              *)
(**********************************************************************)

let reg universe =
  (* Additional output encodings: *)
  universe # application # add_output_encoding "pre1" Definitions.encode_as_pre;

  (* Additional functions: *)
  universe # application # add_var_function "umlmonweb_release"
    (fun _ _ -> String_value Instconf.release);
  universe # application # add_var_function "umlmonweb_revision"
    (fun _ _ -> String_value Instconf.revision);
  
  (* Register dialogs: *)
  universe # register "startpage" (new Startpage.startpage);
  universe # register "vm" (new Vm.vm);
  universe # register "termin" (new Termin.termin);

(*  
  print_error := (extended_error_page universe)
*)
;;
