(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON-Web.
 *
 * UMLMON-Web is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON-Web is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON-Web; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: vm.ml 209 2005-10-23 13:46:46Z gerd $
 * ----------------------------------------------------------------------
 *
 *)
 
open Wd_dialog
open Wd_run_cgi
open Wd_types
open Netcgi_types
open Definitions
 
open Umlmonctrl_aux

open Printf

module Mon = Umlmonctrl_clnt.Ctrl.V1
module Dir = Umlmonctrl_clnt.Directory.V1


let ws_re = Str.regexp "[ \t\r\n]+";;

let del_ctrl_re = Str.regexp "[\000-\031\127]";;

let con_re = Str.regexp "con";;

let ssl_re = Str.regexp "ssl";;

let ubd_cow_re = Str.regexp "^\\([^,]+\\),\\(.+\\)$"


class vm universe name env =
object (self)
  inherit dialog universe name env
  inherit std_behaviour

  method private try_prepare_page() =
    let status = Common.get_status (self # string_variable "id") self#creds in
    self # set_variable "status" (String_value status);
    match self # page_name with
      | "home"        -> self # try_prepare_home()
      | "fn_logfiles" -> self # try_prepare_logfiles()
      | "fn_consoles" -> self # try_prepare_consoles()
      | "fn_disks"    -> self # try_prepare_disks()
      | "fn_config"   -> self # try_prepare_config()
      | _             -> ()

  method private try_prepare_home() =
    (* Just delete variables used by subdialogs *)
    (* --- disks --- *)
    self # set_variable "disk_name" (Alist_value []);
    self # set_variable "disk_device" (Alist_value []);
    self # set_variable "disk_type" (Alist_value []);
    self # set_variable "disk_status" (Alist_value []);
    self # set_variable "disk_size" (Alist_value []);
    self # set_variable "disk_backing" (Alist_value []);
    self # set_variable "disk_error" (Alist_value []);
    self # set_variable "disk_ops" (Alist_value []);
    self # set_variable "disk_selected" (String_value "");
    self # set_variable "disk_new_name" (String_value "");
    self # set_variable "disk_new_size" (String_value "");
    self # set_variable "disk_new_device" (String_value "");
    self # set_variable "disk_base_devices" (Dyn_enum_value []);
    (* --- config --- *)
    self # set_variable "config_name" (Alist_value []);
    self # set_variable "config_value" (Alist_value []);
    self # set_variable "config_rw" (Alist_value []);
    self # set_variable "config_selected" (String_value "");
    self # set_variable "config_new_value" (String_value "");
    self # set_variable "config_selected_create" (String_value "");
    (* omitted: config_base_start!!! *)
    self # set_variable "config_base_create" (Dyn_enum_value []);
    (* --- passwords --- *)
    self # set_variable "pw_old" (String_value "");
    self # set_variable "pw_new1" (String_value "");
    self # set_variable "pw_new2" (String_value "");
    self # set_variable "pw_msg" (String_value "");
    (* --- cloning --- *)
    self # set_variable "clone_msg" (String_value "");
    self # set_variable "clone_new_name" (String_value "");
    (* --- delete --- *)
    self # set_variable "delete_msg" (String_value "");

  method private try_prepare_logfiles() =
    let id = self # string_variable "id" in
    (* Get the list of possible log files *)
    let config = Common.call_proc id self#creds Mon.config_get () in
    let files =
      ( match config.cfg_logfile with
	  | None -> []
	  | Some f -> ["monitor", "UMLMON"]
      ) @
      ( List.map
	  (fun p -> (p.param_name, ("Konsole " ^ p.param_name)))
	  (List.filter
	     (fun p -> String.contains p.param_value ':')
	     (Array.to_list config.cfg_con)
	  )
      ) @
      ( List.map
	  (fun p -> (p.param_name, ("Schnittstelle " ^ p.param_name)))
	  (List.filter
	     (fun p -> String.contains p.param_value ':')
	     (Array.to_list config.cfg_ssl)
	  )
      ) in
    self # set_variable "log_file_list" (Dyn_enum_value files);
    let sel = self # string_variable "log_file_selected" in
    let sel =
      if sel = "" then
	match files with
	  | [] -> ""
	  | (first,_) :: _ -> first
      else
	sel in
    if sel <> "" then (
      let first_line =
	if self # string_variable "log_file_follow" = "yes" then
	  (-20l)
	else
	  match self # string_variable "log_file_pos" with
	    | "" -> (-20l)
	    | n -> 
		if sel <> self # string_variable "log_file_selected_last" then
		  (-20l)
		else
		  Int32.of_string n in
      let result = Common.call_proc id self#creds Mon.logfile_read (sel,first_line,20l) in
      match result with
	| `code_ok r ->
	    let (clean_lines_rev,_) =
	      List.fold_left
		(fun (lines,n) line ->
		   let lines' =
		     (sprintf "%4ld %s" n (Str.global_replace del_ctrl_re "" line)) ::
		       lines in
		   (lines', Int32.succ n))
		([],r.line_start)
		(Array.to_list r.lines) in
	    self # set_variable 
	      "log_file_pos" (String_value (Int32.to_string r.line_start));
	    self # set_variable 
	      "log_file_contents" 
	      (String_value (String.concat "\n" (List.rev clean_lines_rev)));
	    self # set_variable
	      "log_file_selected_last" (String_value sel);
	| `code_error msg -> raise (Common.Remote_error msg)
	| `code_interrupt -> assert false
	| `code_in_progress -> assert false
    )

  method private try_prepare_consoles() =
    let id = self # string_variable "id" in
    (* Get the list of possible consoles: *)
    let terms = 
      Array.to_list(Common.call_proc id self#creds Mon.get_monitored_terminals ()) in
    let terms' =
      List.map
	(fun term ->
	   let name = term.ctrl_termname in
	   let full_name =
	     if Str.string_match con_re name 0 then
	       ("Konsole " ^ name)
	     else
	       if Str.string_match ssl_re name 0 then
		 ("Serielle Schnittstelle " ^ name)
	       else
		 name
	   in
	   let comment =
	     match term.ctrl_termdef with
	       | `ctrl_pty t ->
		   if t.ctrl_connected then "In Benutzung" else "" 
	   in
	   (name, full_name, comment)
	)
	terms in
    let terms'' =
      List.sort (fun (n1,_,_) (n2,_,_) -> Pervasives.compare n1 n2) terms' in
    self # set_variable 
      "console_list"
      (Dyn_enum_value (List.map (fun (name, full_name, _) -> (name, full_name)) terms''));
    self # set_variable
      "console_comments"
      (Dyn_enum_value (List.map (fun (name, _, comments) -> (name, comments)) terms''));
       

  method private try_prepare_disks() =
    let id = self # string_variable "id" in
    let config = Common.call_proc id self#creds Mon.config_get () in
    let disks = Common.call_proc id self#creds Mon.disk_files () in
    let ubd_list = Array.to_list config.cfg_ubd in

    let ext_disks = 
      Array.map
	(fun disk ->
	   try
	     let ubd =
	       List.find
		 (fun ubd ->
		    if Str.string_match ubd_cow_re ubd.param_value 0 then
		      let cow_file = Str.matched_group 1 ubd.param_value in
		      cow_file = disk.disk_name 
		    else
		      ubd.param_value = disk.disk_name
		 )
		 ubd_list in
	     (Some ubd.param_name, disk)
	   with
	       Not_found -> 
		 (None, disk)
	)
	disks in

    Array.sort
      (fun x y ->
	 match (x,y) with
	   | ((Some x_device, x_disk), (Some y_device, y_disk)) ->
	       if x_device = y_device then
		 compare x_disk.disk_name y_disk.disk_name
	       else
		 compare x_device y_device
	   | ((None, x_disk), (None, y_disk)) ->
	       compare x_disk.disk_name y_disk.disk_name
	   | ((None, _), (Some _, _)) ->
	       1
	   | ((Some _, _), (None, _)) ->
	       (-1)
      )
      ext_disks;

    let ext_disks = Array.to_list ext_disks in

    let disk_name =
      List.map 
	(fun (_,disk) -> (disk.disk_name, String_value disk.disk_name)) 
	ext_disks in
    let disk_device =
      List.map
	(fun (dev_opt, disk) ->
	   match dev_opt with
	     | None -> (disk.disk_name, String_value "")
	     | Some dev -> (disk.disk_name, String_value dev))
	ext_disks in
    let disk_status =
      List.map 
	(fun (_,disk) -> 
	   (disk.disk_name,
	    String_value(match disk.disk_info with
			   | `disk_exists info ->
			       if info.disk_use_count > 0l then "BACKING" 
			       else
				 if info.disk_lock = `lock_write then
				   "LOCKED"
				 else
				   if not info.disk_rw then
				     "AVAIL-RO"
				   else
				     ( match info.disk_lock with
					 | `lock_read
					 | `lock_read_own -> "AVAIL-SH"
					 | _ -> "AVAIL-RW"
				     )
			   | `disk_being_created -> "CREATING"
			   | `disk_error _ -> "ERROR")))
	ext_disks in
    let disk_type =
      List.map 
	(fun (_,disk) -> (disk.disk_name, String_value(if disk.disk_is_cow then
							 "COW" else "DSK")))
	ext_disks in
    let disk_size =
      List.map 
	(fun (_,disk) -> 
	   (disk.disk_name,
	    String_value(match disk.disk_info with
			   | `disk_exists info -> 
			       Int64.to_string (Int64.div info.disk_size 1048576L) ^ "M"
			   | _ -> "")))
	ext_disks in
    let disk_backing =
      List.map 
	(fun (_,disk) -> 
	   (disk.disk_name,
	    String_value(match disk.disk_info with
			   | `disk_exists info -> 
			       ( match info.disk_backing_file with
				   | None -> ""
				   | Some bf -> bf)
			   | _ -> "")))
	ext_disks in
    let disk_error =
      List.map 
	(fun (_,disk) -> 
	   (disk.disk_name,
	    String_value(match disk.disk_info with
			   | `disk_error msg -> msg
			   | _ -> "")))
	ext_disks in
    let disk_ops =
      List.map 
	(fun (dev_opt,disk) ->
	   let use_count = 
	     match disk.disk_info with
	       | `disk_exists info -> Int32.to_int info.disk_use_count
	       | _ -> 0 in
	   let exists =
	      match disk.disk_info with
	       | `disk_exists _ -> true
	       | _ -> false in
	   let allow_disk_create_cow = exists && not disk.disk_is_cow && dev_opt = None in
	   let allow_disk_delete = exists && disk.disk_owned && use_count = 0 in
	   let allow_disk_copy = exists in
	   let allow_disk_resize = exists && not disk.disk_is_cow && disk.disk_owned && use_count = 0 in
	   let allow_disk_is_device = exists && disk.disk_owned && use_count = 0 in
	   let s =
	     (if allow_disk_create_cow then "W" else "") ^
	     (if allow_disk_delete then "D" else "") ^
	     (if allow_disk_copy then "C" else "") ^ 
	     (if allow_disk_resize then "R" else "") ^
	     (if allow_disk_is_device then "V" else "") in
	   (disk.disk_name, String_value s))
	ext_disks in

    self # set_variable "disk_name"    (Alist_value disk_name);
    self # set_variable "disk_device"  (Alist_value disk_device);
    self # set_variable "disk_type"    (Alist_value disk_type);
    self # set_variable "disk_ops"     (Alist_value disk_ops);
    self # set_variable "disk_status"  (Alist_value disk_status);
    self # set_variable "disk_size"    (Alist_value disk_size);
    self # set_variable "disk_backing" (Alist_value disk_backing);
    self # set_variable "disk_error"   (Alist_value disk_error);


  method private try_prepare_config() =
    let id = self # string_variable "id" in
    let config = Common.call_proc id self#creds Mon.config_get () in
    let kernels = Common.call_proc id self#creds Mon.announced_kernels() in
    let convert ?(add_suffix=false) param_arr param_rw =
      List.sort
	(fun (x,_,_,_) (y,_,_,_) -> Pervasives.compare x y)
	(Array.to_list
	   (Array.mapi
	      (fun k p -> 
		 let suffix = if add_suffix then string_of_int k else "" in
		 (p.param_name ^ suffix, p.param_name, p.param_value, param_rw)) 
	      param_arr)) in
    let params =
      convert config.cfg_privsetup false @
      convert [| config.cfg_start |] config.cfg_start_rw @
      convert ( match config.cfg_logfile with
		  | None -> [| |] 
		  | Some f -> [| f |] ) config.cfg_logfile_rw @
      convert [| config.cfg_kernel |] config.cfg_kernel_rw @
      convert ~add_suffix:true config.cfg_kernelargs config.cfg_kernelargs_rw @
      convert [| config.cfg_mem |] config.cfg_mem_rw @
      convert config.cfg_con config.cfg_con_rw @
      convert config.cfg_ssl config.cfg_ssl_rw @
      convert config.cfg_eth config.cfg_eth_rw @
      convert config.cfg_extension true
    in
    let config_name =
      List.map (fun (n,nx,v,rw) -> (n, String_value nx)) params in
    let config_value =
      List.map (fun (n,nx,v,rw) -> (n, String_value v)) params in
    let config_rw =
      List.map (fun (n,nx,v,rw) -> (n, String_value (if rw then "yes" else "no"))) params in
    self # set_variable "config_name"  (Alist_value config_name);
    self # set_variable "config_value" (Alist_value config_value);
    self # set_variable "config_rw"    (Alist_value config_rw);
    let base_create =
      (if config.cfg_kernelargs_rw then [ "kernelarg" ] else []) @
      (if config.cfg_con_rw then
	 [ "con0"; "con1"; "con2"; "con3"; "con4"; "con5"; "con6"; "con7";
            "con8"; "con9"; "con10"; "con11"; "con12"; "con13"; "con14"; "con15" ]
       else []) @
      (if config.cfg_ssl_rw then
	 [ "ssl0"; "ssl1"; "ssl2"; "ssl3"; "ssl4"; "ssl5"; "ssl6"; "ssl7";
            "ssl8"; "ssl9"; "ssl10"; "ssl11"; "ssl12"; "ssl13"; "ssl14"; "ssl15" ]
       else []) in
    let base_create =
      List.filter
	(fun name ->
	   not (List.exists (fun (n,_,_,_) -> n = name) params))
	base_create in
    self # set_variable "config_base_create" (Dyn_enum_value (List.map (fun n -> (n,n))
								base_create));
    self # set_variable "config_selected_create" (String_value "");
    let base_kernels =
      List.map
	(fun k ->
	   (k.kernel_file, k.kernel_name)
	)
	(Array.to_list kernels) in
    let current_kernel = config.cfg_kernel.param_value in
    let base_kernels' =
      if base_kernels <> [] && 
	 not (List.mem_assoc current_kernel base_kernels) 
      then
	[ current_kernel, current_kernel ]
      else
	[] in

    self # set_variable "config_base_kernels" 
      (Dyn_enum_value (base_kernels @ base_kernels'));
    

  method private try_handle() =
    let id = self # string_variable "id" in
    match self # event with
      (* -- start/stop -- *)
      | Button "do_start" ->
	  let arg_str = self # string_variable "start_arguments" in
	  let args = Array.of_list(Str.split ws_re arg_str) in
	  Common.call_simple_proc id self#creds Mon.start args;
	  self # show_log()
      | Button "do_shutdown" ->
	  Common.call_simple_proc id self#creds Mon.ctrlaltdel ();
	  self # show_log()
      | Button "do_terminate" ->
	  Common.call_simple_proc id self#creds Mon.sysrq "s";
	  Common.call_simple_proc id self#creds Mon.sysrq "u";
	  Common.call_simple_proc id self#creds Mon.halt ();
	  self # show_log()
      | Button "do_kill" ->
	  Common.call_simple_proc id self#creds Mon.kill ();
	  self # show_log()
      | Button "do_continue" ->
	  Common.call_simple_proc id self#creds Mon.continue ()
      (* -- logfiles -- *)
      | Button "log_file_back" ->
	  self # set_variable "log_file_follow" (String_value "no");
	  let pos = self # string_variable "log_file_pos" in
	  if pos <> "" then
	    self # set_variable 
	      "log_file_pos"
	      (String_value (Int32.to_string (Int32.sub (Int32.of_string pos) 20l)))
      | Button "log_file_fwd" ->
	  self # set_variable "log_file_follow" (String_value "no");
	  let pos = self # string_variable "log_file_pos" in
	  if pos <> "" then
	    self # set_variable 
	      "log_file_pos"
	      (String_value (Int32.to_string (Int32.add (Int32.of_string pos) 20l)))
      | Button "log_file_start" ->
	  self # set_variable "log_file_follow" (String_value "no");
	  self # set_variable "log_file_pos" (String_value "0")
      | Button "log_file_end" ->
	  self # set_variable "log_file_follow" (String_value "no");
	  self # set_variable "log_file_pos" (String_value "-20")
      | Button "log_file_follow" ->
	  self # set_variable "log_file_follow" (String_value "yes");

      (* --- consoles --- *)

      | Indexed_button ("console_select", name) ->
	  self # console_connect name
      | Button("console_change") ->
	  self # console_disconnect()

      (* --- disks --- *)

      | Button "disk_create_cow"
      | Button "disk_copy" ->
	  self # set_variable "disk_new_name" (String_value "")
      | Button "disk_create" ->
	  self # set_variable "disk_new_name" (String_value "");
	  self # set_variable "disk_new_size" (String_value "")
      | Button "disk_resize" ->
	  self # set_variable "disk_new_size" (String_value "")
      | Button "disk_set_device" ->
	  let s = self # string_variable "disk_selected" in
	  let my_dev = 
	    try self # lookup_string_variable "disk_device" s with Not_found -> "" in
	  let used_devs =
	    List.flatten
	      (List.map
		 (function 
		    | (name, String_value dev) ->
			if dev = "" || dev = my_dev then [] else [dev]
		    | _ -> 
			assert false)
		 (self # alist_variable "disk_device")) in
	  let base_devs =
	    List.filter
	      (fun dev -> not (List.mem dev used_devs))
	      [ "ubd0"; "ubd1"; "ubd2"; "ubd3"; "ubd4"; "ubd5"; "ubd6"; "ubd7";
	        "ubd8"; "ubd9"; "ubd10"; "ubd11"; "ubd12"; "ubd13"; "ubd14"; "ubd15" ] in
	  self # set_variable "disk_new_device" (String_value my_dev);
	  self # set_variable "disk_base_devices"
	    (Dyn_enum_value
	       ( ("", "Platte ist keinem Gert zugewiesen") ::
		   (List.map (fun dev -> (dev, "Gert " ^ dev)) base_devs) ))
      | Button "do_disk_create_cow" ->
	  let old_name = self # string_variable "disk_selected" in
	  let new_name = "/disks/" ^ self # string_variable "disk_new_name" ^ ".cow" in
	  Common.call_simple_proc id self#creds Mon.disk_create_cow (old_name,new_name, [| |])
      | Button "do_disk_create" ->
	  let new_name = "/disks/" ^ self # string_variable "disk_new_name" ^ ".dsk" in
	  let new_size = Int32.of_string (self # string_variable "disk_new_size") in
	  Common.call_simple_proc id self#creds Mon.disk_create (new_name, new_size, [| "sparse" |])
      | Button "do_disk_delete" ->
	  let old_name = self # string_variable "disk_selected" in
	  Common.call_simple_proc id self#creds Mon.disk_delete (old_name, [| |]);
	  let config = Common.call_proc id self#creds Mon.config_get () in
	  let disk_params = Array.to_list config.cfg_ubd in
	  let new_disk_params =
	    List.filter
	      (fun p -> p.param_value <> old_name)
	      disk_params in
	  let new_config = { config with 
			       cfg_privsetup = [| |];
			       cfg_ubd = Array.of_list new_disk_params } in
	  Common.call_simple_proc id self#creds Mon.config_set new_config
      | Button "do_disk_copy" ->
	  let old_name = self # string_variable "disk_selected" in
	  let suffix =
	    if Filename.check_suffix old_name ".cow" then ".cow" else ".dsk" in
	  let new_name = "/disks/" ^ self # string_variable "disk_new_name" ^ suffix in
	  Common.call_simple_proc id self#creds Mon.disk_copy (old_name,new_name, [| |])
      | Button "do_disk_resize" ->
	  let old_name = self # string_variable "disk_selected" in
	  let new_size = Int32.of_string (self # string_variable "disk_new_size") in
	  Common.call_simple_proc id self#creds Mon.disk_resize (old_name, new_size, [| "sparse" |])
      | Button "do_disk_set_device" ->
	  let old_name = self # string_variable "disk_selected" in
	  let new_device = self # string_variable "disk_new_device" in
	  let config = Common.call_proc id self#creds Mon.config_get () in
	  let disk_params = Array.to_list config.cfg_ubd in
	  let new_disk_params =
	    ( if new_device = "" then
		[]
	      else
		[ { param_name = new_device; param_value = old_name } ] ) @
	    ( List.filter
	        (fun p ->
		   p.param_name <> new_device && p.param_value <> old_name
	        )
	        disk_params) in
	  let new_config = { config with 
			       cfg_privsetup = [| |];
			       cfg_ubd = Array.of_list new_disk_params } in
	  Common.call_simple_proc id self#creds Mon.config_set new_config

      (* --- config --- *)

      | Button "config_change" ->
	  let selected = self # string_variable "config_selected" in
	  let value = 
	    try self # lookup_string_variable "config_value" selected 
	    with Not_found -> "" in
	  let value' =
	    if selected = "logfile" then
	      Filename.basename value
	    else
	      if Str.string_match (Str.regexp "^\\(con\\|ssl\\)") selected 0 then (
		if Str.string_match (Str.regexp "^pty:") value 0 then
		  Filename.basename (String.sub value 4 (String.length value - 4))
		else
		  ""
	      )
	      else
		value in
          self # set_variable "config_new_value" (String_value value')
      | Button "config_create" ->
          self # set_variable "config_new_value" (String_value "")
      | Button "do_config_change" ->
	  let selected = self # string_variable "config_selected" in
	  self # config_change selected
      | Button "do_config_create" ->
	  let selected = self # string_variable "config_selected_create" in
	  self # config_change selected
      | Button "do_config_delete" ->
	  let selected = self # string_variable "config_selected" in
	  self # config_delete selected

      (* --- passwords --- *)

      | Button "do_pw_change" ->
	  ( try
	      let old_pw = self # string_variable "pw_old" in
	      let new_pw1 = self # string_variable "pw_new1" in
	      let new_pw2 = self # string_variable "pw_new2" in
	      if new_pw1 <> new_pw2 then
		failwith "Eingabefehler: Die neuen Passwrter sind nicht identisch!";
	      if new_pw1 = "" then
		failwith "Eingabefehler: Neues Passwort ist leer!";
	      Common.check_auth ~userpw:true id [old_pw];   (* or Wrong_password *)
	      ( try
		  Common.call_proc id [old_pw] 
		    (fun c -> Umlmon_client.change_password c old_pw)
		    new_pw1
		with
		  | Failure _ ->
		      failwith "Passwort-nderung nicht erfolgreich"
	      );
	      self # set_variable "password" (String_value new_pw1);
	      raise (Change_page "home")
	    with
	      | Common.Wrong_password ->
		  self # set_variable "pw_msg" 
		    (String_value "Das bisherige Passwort ist inkorrekt!")
	      | Failure msg ->
		  self # set_variable "pw_msg" (String_value msg)
	  )

      (* --- cloning --- *)

      | Button "do_clone" ->
	  let new_name = self # string_variable "clone_new_name" in
	  ( match 
	      Common.call_proc id self#creds Mon.site_command [| "clone"; new_name |]
	    with
	      | `code_ok _ ->
		  return_to_caller self
	      | `code_error msg ->
		  self # set_variable "clone_msg" (String_value msg)
	      | _ ->
		  ()
	  )

      (* --- delete --- *)

      | Button "do_delete" ->
	  ( try
	      ( match 
		  Common.call_proc id self#creds Mon.site_command [| "delete" |]
		with
		  | `code_ok _ ->
		      return_to_caller self
		  | `code_error msg ->
		      self # set_variable "clone_msg" (String_value msg)
		  | _ ->
		      ()
	      )
	    with
		Rpc_client.Message_lost ->
		  (* This is very normal! *)
		  return_to_caller self
	  )

      | _ ->
	  ()

  method private show_log() =
    (* Jump to log file: *)
    let id = self # string_variable "id" in
    let config = Common.call_proc id self#creds Mon.config_get () in
    let lf =
      if List.exists (fun p -> p.param_name = "con0") (Array.to_list config.cfg_con) then
	"con0"
      else
	match config.cfg_logfile with
	  | None -> ""
	  | Some f -> "monitor" in
    self # set_variable "log_file_selected" (String_value lf);
    self # set_variable "log_file_selected_last" (String_value "");
    self # set_variable "log_file_pos" (String_value "");
    self # set_variable "log_file_follow" (String_value "yes");
    raise (Change_page "fn_logfiles");

  method private console_connect name =
    let id = self # string_variable "id" in
    match Common.call_proc id self#creds Mon.connect name with
      | `code_ok handle ->
	  self # set_variable 
	    "console_selected_name" (String_value (handle.ct_name));
	  self # set_variable 
	    "console_selected_id" (String_value (Int32.to_string handle.ct_id));
	  Common.call_simple_proc id self#creds Mon.te_init (handle,80l,24l);
      | `code_error msg ->
	  raise (Common.Remote_error msg)
      | `code_interrupt ->
	  raise Common.Interrupted
      | `code_in_progress ->
	  ()

  method private console_disconnect() =
    let id = self # string_variable "id" in
    let name = self # string_variable "console_selected_name" in
    let ct_name = self # string_variable "console_selected_name" in
    let ct_id = self # string_variable "console_selected_id" in
    let handle = { ct_name = ct_name; ct_id = Int32.of_string ct_id } in
    self # set_variable "console_selected_name" (String_value "");
    self # set_variable "console_selected_id" (String_value "");
    ignore(Common.call_proc id self#creds Mon.disconnect (handle,false))

  method private config_change selected =
    let id = self # string_variable "id" in
    let config = Common.call_proc id self#creds Mon.config_get () in
    let value = self # string_variable "config_new_value" in
    let config' =
      match selected with
	| "start" ->
	    { config with
		cfg_start = { param_name = "start"; param_value = value } }
	| "logfile" ->
	    { config with
		cfg_logfile = Some { param_name = "logfile"; param_value = value } }
	| "kernel" ->
	    { config with
		cfg_kernel = { param_name = "kernel"; param_value = value } } 
	| "mem" ->
	    { config with
		cfg_mem = { param_name = "mem"; param_value = value } }
	| "kernelarg" ->  (* Only when parameter is created! *)
	    { config with
		cfg_kernelargs = Array.append config.cfg_kernelargs
				   [| { param_name = "kernelarg"; 
				        param_value = value } |] }
	| _ when Str.string_match (Str.regexp "^kernelarg\\([0-9]+\\)") selected 0 ->
	    let n = int_of_string (Str.matched_group 1 selected) in
	    if n >= Array.length config.cfg_kernelargs then
	      failwith "Parameter does no longer exist";
	    config.cfg_kernelargs.(n) <- { param_name = "kernelarg";
				        param_value = value };
	    config
	| _ when Str.string_match (Str.regexp "^con") selected 0 ->
	    let s = if value = "" then "pty" else "pty:" ^ value in
	    { config with
		cfg_con = Array.of_list
			    ( { param_name = selected; param_value = s } ::
				( List.filter
				    (fun p -> p.param_name <> selected)
				    (Array.to_list config.cfg_con))) }
	| _ when Str.string_match (Str.regexp "^ssl") selected 0 ->
	    let s = if value = "" then "pty" else "pty:" ^ value in
	    { config with
		cfg_ssl = Array.of_list
			    ( { param_name = selected; param_value = s } ::
				( List.filter
				    (fun p -> p.param_name <> selected)
				    (Array.to_list config.cfg_ssl))) }
	| _ -> 
	    let is_extension =
	      List.exists
		(fun p -> p.param_name = selected) 
		(Array.to_list config.cfg_extension) in
	    if is_extension then
	      { config with
		  cfg_extension = (Array.map
				     (fun p -> 
					if p.param_name = selected then
					  { param_name = selected;
					    param_value = value }
					else p)
				     config.cfg_extension) }		  
	    else
	      config
    in
    Common.call_simple_proc id self#creds Mon.config_set { config' with cfg_privsetup = [| |] }


  method private config_delete selected =
    let id = self # string_variable "id" in
    let config = Common.call_proc id self#creds Mon.config_get () in
    let config' =
      match selected with
	| _ when Str.string_match (Str.regexp "^kernelarg\\([0-9]+\\)") selected 0 ->
	    let n = int_of_string (Str.matched_group 1 selected) in
	    if n >= Array.length config.cfg_kernelargs then
	      failwith "Parameter does no longer exist";
	    config.cfg_kernelargs.(n) <- { param_name = "";
				        param_value = "" };
	    { config with
		cfg_kernelargs = Array.of_list
				   (List.filter
				      (fun p -> p.param_name <> "")
				      (Array.to_list config.cfg_kernelargs)) }
	| _ when Str.string_match (Str.regexp "^con") selected 0 ->
	    { config with
		cfg_con = Array.of_list
			    ( List.filter
				(fun p -> p.param_name <> selected)
				(Array.to_list config.cfg_con)) }
	| _ when Str.string_match (Str.regexp "^ssl") selected 0 ->
	    { config with
		cfg_ssl = Array.of_list
			    ( List.filter
				(fun p -> p.param_name <> selected)
				(Array.to_list config.cfg_ssl)) }
	| _ ->
	    let is_extension =
	      List.exists
		(fun p -> p.param_name = selected) 
		(Array.to_list config.cfg_extension) in
	    if is_extension then
	      { config with
		  cfg_extension = (Array.of_list
		                     ( List.filter
					 (fun p -> p.param_name <> selected)
					 (Array.to_list config.cfg_extension)))
	      }
	    else
	      config
    in
    Common.call_simple_proc id self#creds Mon.config_set { config' with cfg_privsetup = [| |] }


  method private check_auth () = ()
end
;;


Registry.new_vm :=
  fun universe env id ->
    let dlg = universe # create env "vm" in
    dlg # set_variable "id" (String_value id);
    dlg # set_variable "nav_name" (String_value ("VM " ^ id));
    dlg
;;
